#include "string.h"
#include "printf.h"
#include "failure.h"
#include "page.h"
#include "multiboot.h"
#include "system.h"
#include "msr.h"
#include "elf.h"
#include "vm.h"
#include "svm.h"
#include "alloc.h"
#include "vmexit.h"
#include "vmm.h"


static struct vmcb *
alloc_vmcb ( void )
{
	struct vmcb *vmcb;

	const unsigned long pfn = alloc_pages ( 1, 1 );
	vmcb = ( struct vmcb * ) VIRT ( pfn << PAGE_SHIFT );
	memset ( ( char * ) vmcb, 0, sizeof ( struct vmcb ) );

	return vmcb;
}

static unsigned long 
create_intercept_table ( unsigned long size )
{
	const unsigned long pfn = alloc_pages ( size >> PAGE_SHIFT, 1 );
	void *p = ( void * ) VIRT ( pfn << PAGE_SHIFT );

	/* vol. 2, p. 445 */
	memset ( p, 0xff, size );

	return pfn << PAGE_SHIFT; 
}

static void	
set_control_area ( struct vmcb *vmcb )
{
#if 0
	/* Enable nested paging (See AMD64 manual Vol. 3, p. 488) */
	vmcb->np_enable = 1; 
	printf ( "Nested paging enabled.\n" );
#endif

	/* Flush TLB on VMRUN (all entries, all ASIDs) */
        vmcb->tlb_control = 1;

	/* To be added in RDTSC and RDTSCP */
	vmcb->tsc_offset = 0; 
	
	/* Guest address space identifier (ASID) */
	vmcb->guest_asid = 1;

	/* Intecept the VMRUN instruction */
	vmcb->general2_intercepts = INTRCPT_VMRUN;

	/* [REF] vol.2, p. 454 */
	vmcb->iopm_base_pa  = create_intercept_table ( 12 << 10 ); /* 12 Kbytes */
	vmcb->msrpm_base_pa = create_intercept_table ( 8 << 10 );  /* 8 Kbytes */
}

/* Setup the segment registers and all their hidden states 
 * (See AMD64 manual Vol 2, pp. 98-104 to know the format of generic segment descriptors) */
static void
set_descriptors ( struct vmcb *vmcb )
{
	int i;

	/* Must be a 32-bit read/execute code segment with an offset of 0 and a limit of 0xffffffff. 
	 * The exact value is undefined. */
	vmcb->cs.sel         = 1 << 3; /* dummy */
	vmcb->cs.attrs.bytes = 0x0c9a; // 1100 1001 1010 = Code-segment, R(eadable), S(ystem/user), and P(resent).
//	vmcb->cs.attrs.bytes = 0x0c1a; // 1100 1001 1010 = Code-segment, R(eadable), S(ystem/user), and P(resent).
	vmcb->cs.limit       = 0xfffff;
//	vmcb->cs.limit       = 0x80000;
	vmcb->cs.base        = 0;

	/* Must be a 32-bit read/write data segment with an offset of 0 and a limit of 0xffffffff. 
	 * The exact values are all undefined. */
	struct seg_selector *segregs [] = { &vmcb->ss, &vmcb->ds, &vmcb->es, &vmcb->fs, &vmcb->gs, NULL };
	for ( i = 0; segregs [ i ] != NULL; i++ ) {
		struct seg_selector *x = segregs [ i ];
		x->sel         = 0 << 3; /* dummy */
		x->attrs.bytes = 0x0c92; // 0000 1001 0010 = Data-segment, W(ritable), S(ystem/user), and P(resent). 
//		x->attrs.bytes = 0x0c12; // 1100 1001 0010 = Data-segment, W(ritable), S(ystem/user), and P(resent). 
		x->limit       = 0xfffff;
//		x->limit = 0x80000;
		x->base        = 0;
	}

	/* The GDTR and IDTR may be invalid.
	 * The OS image must not load any segment registers until it sets up its own GDT. 
	 * The OS image must leave interrupts disabled until it sets up its own IDT */
	struct seg_selector *dsegs [] = { &vmcb->idtr, &vmcb->ldtr, &vmcb->tr, NULL };
	for ( i = 0; dsegs [ i ] != NULL; i++ ) {
		struct seg_selector *x = dsegs [ i ];
		x->sel         = 0;
		x->attrs.bytes = 0;
		x->limit       = 0;
		x->base        = 0;
	}

	/* dummy */
	vmcb->gdtr.limit = 2 * 8 - 1;
	vmcb->gdtr.base  = 0;
}

/* See Multiboot Specification Manual */
static void
set_state_save_area ( struct vmcb *vmcb )
{
	vmcb->cpl = 0; /* must be equal to SS.DPL  */

	/* Must contatin the magin value */
	vmcb->rax = MULTIBOOT_BOOTLOADER_MAGIC;

	/* The OS image must create its own stack as soon as it needs one. */
	vmcb->rsp = 0;

	/* Bit 31 (PG) must be cleared.  Bit 0 (PE) must be set. Other bits are all undefined. */
	vmcb->cr0 = X86_CR0_PE; 
	vmcb->cr3 = 0;
	vmcb->cr4 = 0;

	/* Bit 17 (VM) must be cleared. Bit 9 (IF) must be cleared.  Other bits are all undefined */
	vmcb->rflags = 2; /* set the reserved bit */

#if 0
	vmcb->g_pat = 0x7040600070406UL;	
#else
	{
		u32 eax, edx;
		rdmsr ( MSR_PAT, eax, edx );
		vmcb->g_pat = ( ( ( (u64) edx ) << 32 ) | eax );
	}
#endif
	
	vmcb->efer = EFER_SVME;

	set_descriptors ( vmcb );

	/* [TODO] A20 gate must be enabled. */
}

static unsigned long
alloc_vm_pmem ( unsigned long size )
{
	const unsigned long align = 1 << ( PAGE_SHIFT_2MB - PAGE_SHIFT ); // 2 MB page table ˤ碌 alignment ɬ
	const unsigned long pfn   = alloc_pages ( size >> PAGE_SHIFT, align );
	return ( unsigned long ) VIRT ( pfn << PAGE_SHIFT );
}

static struct multiboot_info *
init_vm_mbi ( unsigned long vm_pmem_start )
{
	enum { INSTALL_PADDR = 0x2d0e0UL }; /* < 1 MB [TODO] */
	struct multiboot_info *mbi = ( struct multiboot_info *) ( vm_pmem_start + INSTALL_PADDR );

	/* [TODO] */

	printf ( "Multiboot information initialized.\n" );

	return mbi;
}

static int
is_reserved_pmem ( unsigned long pfn )
{
//	return ( pfn < ( 1 << ( 21 - PAGE_SHIFT_2MB ) ) );
	return ( pfn == 0 );
}

/* ۥޥʪ¥ޥʪбŤڡơ֥
 * ʪɥ쥹֤*/
static unsigned long 
create_vm_pmem_mapping_table ( unsigned long vm_pmem_start, unsigned long vm_pmem_size )
{
	const unsigned long cr3  = pml4_table_create ( );
	const unsigned long pml4 = ( unsigned long ) VIRT ( cr3 );
	const unsigned long vm_pmem_pfn = PFN_DOWN_2MB ( PHYS ( vm_pmem_start ) );
	int i;

	for ( i = 0; i < PFN_UP_2MB ( vm_pmem_size ); i++ ) {
		const unsigned long vm_paddr = i << PAGE_SHIFT_2MB;
		const unsigned long pm_pfn   = i + ( is_reserved_pmem ( i ) ? 0 : vm_pmem_pfn ); /* [NAIVE] for VGA */
		const unsigned long pm_paddr = pm_pfn << PAGE_SHIFT_2MB;

		mmap ( pml4, vm_paddr, pm_paddr, 1 /* is_user */ );
	}

	printf ( "Page table for nested paging created.\n" );
	return cr3;
}


/*  Current Stauts
 *
 *  Paging ͭˤƤʤΤˡpage fault ȯƤޤ
 * 
 */

void
vm_create ( struct vm *vm, unsigned long guest_image_start, unsigned long guest_image_size, unsigned long vm_pmem_size )
{
	struct vmcb *vmcb;

	/* Allocate a new page for storing VMCB.  */
	vmcb = alloc_vmcb ( );
	vm->vmcb = vmcb;

	set_control_area ( vm->vmcb );
	set_state_save_area ( vm->vmcb );

	/* Allocate new pages for physical memory of the guest OS.  */
	const unsigned long vm_pmem_start = alloc_vm_pmem ( vm_pmem_size );

	/* Set Host-level CR3 to use for nested paging.  */
	vm->h_cr3   = create_vm_pmem_mapping_table ( vm_pmem_start, vm_pmem_size );
//	vmcb->h_cr3 = vm->h_cr3;

	printf ( "mapping : 0 --> %x\n", vm_pmem_start );

	/* Copy the OS image to the specified region by interpreting the ELF format.  */
	vm->vmcb->rip = load_elf_image ( guest_image_start, guest_image_size, vm_pmem_start );

	/* Setup multiboot info.  */
	vm->mbi = init_vm_mbi ( vm_pmem_start );

	printf ( "RIP = %x\n", vm->vmcb->rip );
	printf ( "New virtual machine created.\n" ); 

	print_pg_table ( VIRT ( vm->h_cr3 ) );  

	// gdt table 򲾤˺
	printf ( "create dummy gdt.\n" );

	// data
	* ( ( u32 * ) ( vm_pmem_start +  0 ) ) = 0xffff;
	* ( ( u32 * ) ( vm_pmem_start +  4 ) ) = ( 0x92 << 8 ) + (0xf << 16) + (0xc << 20);

//	* ( ( u32 * ) ( vm_pmem_start +  0 ) ) = 0;
//	* ( ( u32 * ) ( vm_pmem_start +  4 ) ) = ( 0x12 << 8 ) + (0x8 << 16) + (0xc << 20);

	// code 
	* ( ( u32 * ) ( vm_pmem_start +  8 ) ) = 0xffff;
	* ( ( u32 * ) ( vm_pmem_start + 12 ) ) = ( 0x9a << 8 ) + (0xf << 16) + (0xc << 20);
//	* ( ( u32 * ) ( vm_pmem_start +  8 ) ) = 0;
//	* ( ( u32 * ) ( vm_pmem_start + 12 ) ) = ( 0x1a << 8 ) + (0x8 << 16) + (0xc << 20);

#if 0
	{
		int i;
		
		memset ( vm_pmem_start, 0, PAGE_SIZE );

		for ( i = 0; i < 4; i++ ) {
			mmap ( vm_pmem_start, i << PAGE_SHIFT_2MB, i << PAGE_SHIFT_2MB, 1 /* is_user */ );
		}
	}
#endif
}

/******************************************************/

static void
switch_to_guest_os ( struct vm *vm )
{
	u64 p_vmcb = PHYS ( vm->vmcb );
	svm_launch ( p_vmcb );
}

static void
handle_vmexit ( struct vm *vm )
{
	print_vmexit_exitcode ( vm->vmcb->exitcode );
	printf ( "VMCB: rip=%x\n", vm->vmcb->rip );

	printf ( "cpl=%x\n", vm->vmcb->cpl );
	printf ( "cr0=%x, cr3=%x, cr4=%x\n", vm->vmcb->cr0, vm->vmcb->cr3, vm->vmcb->cr4 );
	printf ( "rflags=%x, efer=%x\n", vm->vmcb->rflags, vm->vmcb->efer );

	printf ( "cs.attrs=%x, ds.attrs=%x\n", vm->vmcb->cs.attrs.bytes, vm->vmcb->ds.attrs.bytes );

	printf ( "error_code=%x, fault address=%x\n", vm->vmcb->exitinfo1, vm->vmcb->exitinfo2 );

	/* p. 268, p.490 */
	if ( vm->vmcb->exitinfo1 & 1 ) {
		printf ( "page fault was caused by a page-protection violation\n" );
	} else {
		printf ( "page fault was caused by a not-present page\n" );
	}

	if ( vm->vmcb->exitinfo1 & 2 ) {
		printf ( "memory access was write\n" );
	} else {
		printf ( "memory access was read\n" );
	}

	if ( vm->vmcb->exitinfo1 & 4 ) {
		printf ( "an access in user mode caused the page fault\n" );
	} else {
		printf ( "an access in supervisor mode caused the page fault\n" );
	}
}

void
vm_boot ( struct vm *vm )
{
	printf ( "Booting guest operating system...\n\n" ); 

	vmcb_check_consistency ( vm->vmcb );

	/* [TODO] setup registers (set %ebx to mbi address) */

	{
		const char code[] = 
			{ 
				0xb8, 0x08, 0x00, 0x00, 0x00, // mov $0x8, %eax
				0x8e, 0xc8,                   // movl %eax, %cs
			  
				0xb8, 0x00, 0x00, 0x00, 0x00, // mov $0x0, %eax
				0x8e, 0xd8                    // movl %eax, %ds
			};
		
		vm->vmcb->rip = 0x30;
		memmove ( (char *)vm->vmcb->rip, code, sizeof ( code ) );

//		printf ( "[%x] vm->vmcb->rip = %x\n", i, vm->vmcb->rip );

		switch_to_guest_os ( vm );

		handle_vmexit ( vm );
	}
	
	printf ( "Not found\n" );   
}
